// Four digit LCD clock

// Current consumption averages 405nA total, giving about 60 years on a 210mAh
// CR2032 if cell self discharge rate & shelf life are ignored.

// Power supply is 1.8v from MCP1820T-18 20nA Iq LDO, plus 3v lithium button cell.
// These are now EOL, TI do the TPS7A02 series which are 25nA Iq.

// The LCD display is DE126-RS-20/7.5, see below for connections.

// PIC VLCD3/RB3 should be tied to Vdd: this avoids use of the internal
// resistive ladder voltage divider which is expensive in terms of power
// budget.

// RTC is MicroCrystal RV-3028-C7, +/-1ppm 45nA @ 3.0v. Using an external RTC
// such as this avoids the use of the relatively power hungry oscillators on the
// PIC during sleep. It also supplies the PIC with a much lower frequency to 
// drive the LCD peripheral, saving even more power by reducing the speed 
// in frequency dividers.

// I2C is bit banged in software, SCL is on RC1, SDA is on RD0. 10k x 2 to Vdd.
// This controls the RTC registers.

// When not in sleep, CPU runs at 16MHz from HFINTOSC except for lengthy delays 
// when it switches to LFINTOSC. Running at 16MHz gives the best uA/MHz.

// The RTC is configured to generate 2048Hz on its push-pull CLKOUT pin, and
// that feeds into T1CKI/RC0.

// The T1CKI 2048Hz is used by both the LCD peripheral and Timer 1, with both
// peripherals operating during sleep.

// 2048Hz is the lowest that the LCD peripheral can go without showing flickering
// using the minimum clock division available on the peripheral: LCD runs at 16Hz.

// RTC Vbak is either a 10k to Vss if not used or a low leakage cap.
// To enable the trickle charge & Vbak, #define ENABLE_VBAK_TRICKLE_CHARGE 
// Using Vbak takes about an additional 45nA @ 1.8v: this is due to
// the low voltage detection in the RTC.

// By default, the RTC refreshes itself by default at midnight, 
// clearing some configurations such as the CLKOUT 2048Hz. This is avoided by 
// clearing the RTC's EERD bit.

// On reset, the PIC checks with the RTC PORF bit to see if the RTC contents are OK.
// If so, it simply reads the time from the RTC. If not, it reconfigures the RTC,
// and used the build time as the default time.

// Timer1 is configured to timeout every minute to update the display.
// The active time to update the display depends on how many digits need 
// updating, and is about ~30us average with -O0 and ~20us with -O3. This gives 
// a duty cycle of about 2000000:1. Current draw for 30us at 16MHz is 
// about 500uA, or an aggregated 250pA with the 2000000:1 duty cycle. For a 
// 6-digit HH:MM:SS clock, this amounts to an additional 15nA aggregated.

// When Timer1 times out, it does not go to the RTC to determine the time, 
// it uses its own RAM copy: to receive the time from the RTC over I2C takes 
// about 1ms, which is expensive in terms of CPU cycles, and power budget, 
// considering the 10k pull ups on I2C lines as well as CPU as it's a software
// I2C implementation. While this is not a big deal every minute (about 8nA
// aggregated over time), on a 1Hz  6-digit HH:MM:SS clock this would add a 
// significant 500uA to the power budget.

// Push buttons:
// Button to Vss on RB1 is down
// Button to Vss on RB5 is up
// Button to Vss on RE3 selects hours for up/down, otherwise minutes

// The push button interface has three buttons. The up/down buttons are on IOC
// pins. When pushed, the processor comes out of sleep and enters a time setting
// mode. If no buttons are pushed for about three seconds, it goes back into 
// sleep mode. The cursor flashes rapidly when it's in this setting mode.

// The time on the display is set & written to the RTC once the colon stops
// flashing: this also resets the seconds to zero.

// During time setting mode, the device runs at 16MHz for the actual 
// processing, but for most of the time it runs on  LFINTOSC while delaying to 
// poll and debounce the buttons. Most of the current draw is from the weak pull 
// ups as the buttons are pressed. If necessary this could be reduced by 
// switching to, say, 1Mohm external pull ups instead of the internal weak pull 
// ups.

// LCD connections

// Note that as the Hours digit only holds the values 0, 1, or 2, we can be
// crafty with the segments, tying 1A, 1D and 1E together using only one
// segment pin. See "RA0/SEG12*" below.

// LCD connections
// LCD pin  PIC pin   
// 01 COM   37 RB4/COM0   
// 02 1E    02 RA0/SEG12*
// 03 1D    02 RA0/SEG12*
// 04 1C    03 RA1/SEG7
// 05 1P    N/C
// 06 2E    05 RA3/SEG15
// 07 2D    06 RA4/SEG4
// 08 2C    07 RA5/SEG5
// 09 2P    N/C
// 10 3E    08 RE0/SEG21
// 11 3D    09 RE1/SEG22
// 12 3C    10 RE2/SEG23
// 13 3P    N/C
// 14 4E    14 RA6/SEG1
// 15 4D    17 RC2/SEG3
// 16 4C    18 RC3/SEG6
// 17 4P    N/C
// 18 5E    N/C
// 19 5D    N/C
// 20 5C    N/C
// 21 5P    N/C
// 22 6E    N/C
// 23 6D    N/C
// 24 6C    N/C
// 25 6B    N/C
// 26 6A    N/C
// 27 6F    N/C
// 28 6G    N/C
// 29 5B    N/C
// 30 5A    N/C
// 31 5F    N/C
// 32 5G    N/C
// 33 COL2  N/C
// 34 4B    35 RB2/SEG25
// 35 4A    21 RD2/SEG28
// 36 4F    22 RD3/SEG16
// 37 4G    23 RC4/SEG11
// 38 3B    24 RC5/SEG10
// 39 3A    25 RC6/SEG9
// 40 3F    26 RC7/SEG8
// 41 3G    27 RD4/SEG17
// 42 COL1  20 RD1/SEG27
// 43 2B    28 RD5/SEG18
// 44 2A    29 RD6/SEG19
// 45 2F    30 RD7/SEG20
// 46 2G    33 RB0/SEG0
// 47 1B    40 RB7/SEG13
// 48 1A    02 RA0/SEG12*
// 49 1F    13 RA7/SEG2
// 50 1G    39 RB6/SEG14

// PIC16LF1907 Configuration Bit Settings

// 'C' source line config statements

// CONFIG1
#pragma config FOSC = INTOSC    // Oscillator Selection (INTOSC oscillator: I/O function on CLKIN pin)
#pragma config WDTE = OFF       // Watchdog Timer Enable (WDT disabled)
#ifdef __DEBUG
    #pragma config PWRTE = OFF      // Power-up Timer Enable (PWRT disabled)
    #pragma config MCLRE = ON       // MCLR Pin Function Select (MCLR/VPP pin function is MCLR)
#else
    #pragma config PWRTE = ON       // Power-up Timer Enable (PWRT enabled)
    #pragma config MCLRE = OFF      // MCLR Pin Function Select (MCLR/VPP pin function is digital input)
#endif
#pragma config CP = OFF         // Flash Program Memory Code Protection (Program memory code protection is disabled)
#pragma config BOREN = OFF      // Brown-out Reset Enable (Brown-out Reset disabled)
//#pragma config CLKOUTEN = ON    // Clock Out Enable (CLKOUT function is enabled on the CLKOUT pin)
#pragma config CLKOUTEN = OFF   // Clock Out Enable (CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin)

// CONFIG2
#pragma config WRT = OFF        // Flash Memory Self-Write Protection (Write protection off)
#pragma config STVREN = ON      // Stack Overflow/Underflow Reset Enable (Stack Overflow or Underflow will cause a Reset)
#pragma config BORV = LO        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (Vbor), low trip point selected.)
#pragma config LPBOR = OFF      // Low-Power Brown-out Reset (Low-Power BOR is disabled)
//#pragma config LVP = ON         // Low-Voltage Programming Enable (Low-voltage programming enabled)
#pragma config LVP = OFF        // Low-Voltage Programming Enable (High-voltage on MCLR/VPP must be used for programming)

// #pragma config statements should precede project file includes.
// Use project enums instead of #define for ON and OFF.

#include <xc.h>
#include <stdint.h>
#include <stdbool.h>
#include "lcd.h"

//#define ENABLE_VBAK_TRICKLE_CHARGE // Adds ~45nA @ 1.8V. If disabled, tie Vbsk to Vss with a 10k resistor.

#define _XTAL_FREQ (16000000UL)

#define SDA_TRIS TRISDbits.TRISD0
#define SCL_TRIS TRISCbits.TRISC1
#define SDA_LAT LATDbits.LATD0
#define SCL_LAT LATCbits.LATC1
#define SDA_PORT PORTDbits.RD0

#define SDA_DIGITAL()
#define SCL_DIGITAL()
#define SDA_HIGH() SDA_TRIS=1
#define SDA_LOW() SDA_TRIS=0
#define SDA_INPUT() SDA_TRIS=1
#define SDA_OUTPUT() SDA_TRIS=0
#define SCL_HIGH() SCL_TRIS=1
#define SCL_LOW() SCL_TRIS=0
#define SCL_INPUT() SCL_TRIS=1
#define SCL_OUTPUT() SCL_TRIS=0

// Buttons...
// Button to Vss on RB1 is down
// Button to Vss on RB5 is up
// Button to Vss on RE3 selects hours, otherwise minutes
#define BTN_HOURS PORTEbits.RE3

#define FLASH_ROWSIZE 32            // size of a row
#define FLASH_ROWMASK     FLASH_ROWSIZE-1
#if defined(_16LF1907)
    #define HEF_START   0x1F80      // first address in HE FLash memory
    #define HEF_END     0x1FFF      // last address in HE Flash memory
#endif
#define HEF_SIZE (HEF_END-HEF_START+1)

#define I2C_RV3028C7_ADDR (0x52)

// The TIMER1 delay is measured for 1024Hz clock (2048Hz from T1CLKI/RTC CLKOUT plus 1:2 prescaler).
#define TIMER1_DELAY (1024U*60U) // 1 minute
// #define TIMER1_DELAY (1024U) // 1 second, for testing

#ifdef __DEBUG
    #define FORCE_RTC_INIT true
#else
    #define FORCE_RTC_INIT false
#endif  

static const char _szBuildTime[]=__TIME__; // Used to reset the RTC at first time boot, format "HH:MM:SS"
//static const char _szBuildTime[]="11:11:11"; // Used to reset the RTC at first time boot, format "HH:MM:SS"
//static const char _szBuildTime[]="01:22:22"; // Used to reset the RTC at first time boot, format "HH:MM:SS": 01:22 is 18 segs, average for 4 digit display; 01:22:22 is 28 segs, average for 6 digit display

typedef struct
{
    int8_t n8Hours10;
    int8_t n8Hours1;
    int8_t n8Mins10;
    int8_t n8Mins1;
} CLKSTRUCT;

static CLKSTRUCT _cs;

typedef union
{
    uint8_t u8;
    struct 
    {
        unsigned b0                    :1;
        unsigned b1                    :1;
        unsigned b2                    :1;
        unsigned b3                    :1;
        unsigned b4                    :1;
        unsigned b5                    :1;
        unsigned b6                    :1;
        unsigned b7                    :1;
    } s;
} U8UNION;

// Button debouncing
static uint8_t _u8BtnPortBState;
static uint8_t _u8BtnPortBStateLast;

// Flash bits and pieces
static uint8_t _au8FlashBuffer[FLASH_ROWSIZE]; // RAM buffer for HEF operations
volatile const uint8_t _au8HEF[HEF_SIZE] __at(HEF_START)={0xFF}; // Reserve and initialize HEF: first byte is either 0xFF (indicating a fresh chip, just been programmed) or 0x00.

static void FlashReadBlock(uint16_t u16Address)
{
    int i;
    uint8_t *pu8=_au8FlashBuffer;
    
    for (i=0;i<FLASH_ROWSIZE;i++)
    {
        PMADR=u16Address;
        PMCON1bits.CFGS=0; // 0 => Program memory, not config memory
        PMCON1bits.RD=1; // 1 => Initiate read
        NOP();
        NOP();

        *pu8++=PMDATL;
        u16Address++;
    }
}

static inline void FlashUnlock(void)
{
    PMCON2=0x55;
    PMCON2=0xAA;
    PMCON1bits.WR=1;
    NOP();
    NOP();
}

static void FlashEraseBlock(uint16_t u16Address)
{
    bool bTemp = INTCONbits.GIE;
    INTCONbits.GIE = 0;
    
    PMADR=u16Address;
    PMCON1bits.CFGS=0; // 0 => Program memory, not config memory
    PMCON1bits.FREE=1; // Erase operation
    PMCON1bits.WREN=1; // Enable writes
    FlashUnlock();
    PMCON1bits.WREN=0; // Disable writes
    
    INTCONbits.GIE=bTemp;
}

static void FlashWriteBlock(uint16_t u16Address)
{
    int i;
    uint8_t *pu8=_au8FlashBuffer;
    bool bTemp = INTCONbits.GIE;
    
    INTCONbits.GIE = 0;

    PMADR=u16Address;
    PMCON1bits.CFGS=0; // 0 => Program memory, not config memory
    PMCON1bits.FREE=0; // 0 => Write operation
    PMCON1bits.LWLO=1; // 1 => Load write latches only, 0 => Write to flash
    PMCON1bits.WREN=1; // 1 => Enable writes

    for (i=0;i<FLASH_ROWSIZE;i++)
    {
        PMDAT=*pu8++;
        if (i>=FLASH_ROWSIZE-1)
        {
            PMCON1bits.LWLO=0; // 1 => Load write latches only, 0 => Write to flash
        }
        FlashUnlock();
        PMADR++;
    }
    PMCON1bits.WREN=0; // Disable writes
    INTCONbits.GIE=bTemp;
}

static void HEFClearFirstByte(void)
{
    // The first byte of HEF, if 0xFF, means it's the first time since this device has been run since being programmed.
    FlashReadBlock(HEF_START); // Read the first HEF block into the RAM buffer
    _au8FlashBuffer[0]=0; // Clear the fist byte of the RAM buffer
    FlashWriteBlock(HEF_START); // Rewrite first HEF block: Note this doesn't need an erase first as we're only changing from all ones to zeros.
}

static void I2CStart(void)
{
    SDA_HIGH();
    SCL_HIGH();
    __delay_us(3);
    SDA_LOW();
    __delay_us(3);
    SCL_LOW();                        // pull SCL low
}

static void I2CStop(void)
{
    SCL_LOW();                        // ensure SCL is low
    __delay_us(3);
    SDA_LOW();                        // SDA low
    __delay_us(3);
    SCL_HIGH();                        // pull SCL high
    __delay_us(3);
    SDA_HIGH();                   // allow SDA to be pulled high
    __delay_us(3);
}

static void I2CWrite(uint8_t u8Data)
{
    int8_t i=8;                // loop counter

    do         // loop through each bit
    {
        if (((U8UNION)u8Data).s.b7)
        {
            SDA_HIGH();
        }
        else
        {
            SDA_LOW();
        }
        // Data setup time is 250ns min
        SCL_HIGH();             // pull SCL high to clock bit
        u8Data <<= 1;           // shift left for next bit - Do it here to avoid additional delays when SCL=0
        NOP(); // Pad out SCL high to min 4us
        NOP(); // Pad out SCL high to min 4us
        SCL_LOW();                        // pull SCL low for next bit            
    }
    while (--i!=0); // Hopefully doing it this way XC8 will use DECFSZ instruction
    
    // Check for ACK
    SDA_HIGH();                    // configure SDA as an input
    // No need for delay @ Fosc=4MHz
    SCL_HIGH();                        // bring SCL high to begin transfer
    while (SDA_PORT) // Wait for ACK
    {
    }
    SCL_LOW();                        // bring SCL low again.    
}

static uint8_t I2CRead(bool bNack)
{
    int8_t i=8;
    uint8_t u8Data=0;

    SDA_HIGH();                   // configure SDA as an input

    do 
    {
        u8Data <<= 1;
        SCL_HIGH();                        // bring SCL high to begin transfer
        if (SDA_PORT)
        {
            u8Data|=1;
        }
        SCL_LOW();                        // bring SCL low again.     
    }
    while (--i!=0);

    if (bNack)
    {
        SDA_HIGH();
    }
    else
    {
        SDA_LOW();
    }
    // Data setup time is 250ns min
    SCL_HIGH();                       // pull SCL high to clock bit
    __delay_us(3); 
    SCL_LOW();                        // pull SCL low for next bit
    return u8Data;
}

static void I2CWriteRegister(uint8_t u8I2CAddr,uint8_t u8RegAddr,uint8_t u8Data)
{
	I2CStart();
	
	I2CWrite((uint8_t)(u8I2CAddr << 1) | 0U);
	I2CWrite(u8RegAddr);
	I2CWrite(u8Data);
	I2CStop();
}

uint8_t I2CReadRegister(uint8_t u8I2CAddr,uint8_t u8RegAddr)
{
  	uint8_t u8Data;
  
	I2CStart();
	
	I2CWrite((uint8_t)(u8I2CAddr << 1) | 0U);
	I2CWrite(u8RegAddr);
	I2CStart();
	I2CWrite((uint8_t)(u8I2CAddr << 1) | 1U);
	u8Data=I2CRead(true);
	I2CStop();
            
   	return u8Data;
} 

static void I2CInit(void)
{
    // I2C bit bang uses GPIOs in simulated open drain mode
    SDA_DIGITAL();
    SCL_DIGITAL();
    SDA_INPUT();
    SCL_INPUT();
    SDA_LAT=0;
    SCL_LAT=0;
#ifdef SDA_ANSEL
    SDA_ANSEL=0;
#endif    
#ifdef SCL_ANSEL
    SCL_ANSEL=0;
#endif 
}

static uint8_t AsciiBCDToPackedBCD(const char *psz) // Helper function when converting __TIME__ string to BCD
{
    uint8_t u8=((uint8_t)psz[1]) & 0x0F;
    
    u8 |= (((uint8_t)psz[0])<<4);
    
    return u8;
}

static void LPDelayMS(uint8_t u8Delay)
{
    // Low power delay: switch to LFINTOSC 31kHz
    // Note that Timer 0 will not operate in sleep so we have to poll
    // 255 >= u8Delay >= 1
    // delay[ms] = 2.55 + (0.956 * u8MX) empirically approx, or:
    // u8Delay = (delay[ms] - 2.55) / 0.956 approx
    
    uint8_t u8IRCFSave=OSCCONbits.IRCF;
    
    OPTION_REGbits.TMR0CS=0; // 0 => Fosc/4
    OPTION_REGbits.PS=0b010; // 0b010 => 1:8 timer 0 prescaler ( @ 31kHz Fosc, ~= 1ms [ Fosc/4/8] )
    OPTION_REGbits.PSA=0; // 0 => Prescaler is assigned to the Timer0 module
    INTCONbits.TMR0IE=0;
    OSCCONbits.IRCF=0b0000; // 0b0000 => LFINTOSC
    while (!OSCSTATbits.LFIOFR) // Wait for LFINTOSC to be ready
    {
    }
    TMR0=(uint8_t)(256-u8Delay);
    INTCONbits.TMR0IF=0;
    while (!TMR0IF)
    {
    }
    
    OSCCONbits.IRCF=u8IRCFSave;
    OPTION_REGbits.TMR0CS=1; // 1 => T0CKI (POR default))
    OPTION_REGbits.PS=0b11; // 0b111 => 1:256 timer 0 prescaler (POR default)
    OPTION_REGbits.PSA=1; // 0 => Prescaler is not assigned to the Timer0 module (default)
    TMR0=0;
}

static void RTCWriteTime(void) // Write time from RAM to RTC
{
    uint8_t u8;
    
    I2CWriteRegister(I2C_RV3028C7_ADDR,0x00,0); // Seconds

    u8=(uint8_t)(_cs.n8Mins10<<4) | (uint8_t)_cs.n8Mins1;
    I2CWriteRegister(I2C_RV3028C7_ADDR,0x01,u8); // Minutes

    u8=(uint8_t)(_cs.n8Hours10<<4) | (uint8_t)_cs.n8Hours1;
    I2CWriteRegister(I2C_RV3028C7_ADDR,0x02,u8); // Hours
}

static void RTCReadTime(void) // Read time from RTC to RAM
{
    uint8_t u8;
    
    u8=I2CReadRegister(I2C_RV3028C7_ADDR,0x01); // Minutes
    _cs.n8Mins10 = u8 >> 4;
    _cs.n8Mins1 = u8 & 0x0F;

    u8=I2CReadRegister(I2C_RV3028C7_ADDR,0x02); // Hours
    _cs.n8Hours10 = u8 >> 4;
    _cs.n8Hours1 = u8 & 0x0F;  
}

static void RTCInit(void)
{
    uint8_t u8RTCStatus;

    // Wait for RTC's EEbusy to go inactive
    {
        bool bWaitingForRTC=true;

        while (bWaitingForRTC)
        {
            u8RTCStatus=I2CReadRegister(I2C_RV3028C7_ADDR,0x0E); // Read RTC status register

            bWaitingForRTC=(u8RTCStatus & 0x80)!=0; // Check EEbusy bit
            if (bWaitingForRTC)
            {
                LPDelayMS(50); // Low power delay
            }
        }
    }

    // Write the build date & time to the RTCC if: RTC says POR, or FORCE_RTC_INIT macro says so (eg when debugging), or we're newly programmed (first byte of HEF==0xFF)
    if ((u8RTCStatus & 0x01)!=0 || FORCE_RTC_INIT || _au8HEF[0]==0xFF)
    {
        {
            uint8_t u8;

            // Convert __TIME__ format HH:MM:SS to BCD digits, one nibble per byte, into RAM clock
            u8=AsciiBCDToPackedBCD(_szBuildTime+3);
            _cs.n8Mins10 = u8 >> 4;
            _cs.n8Mins1 = u8 & 0x0F;

            u8=AsciiBCDToPackedBCD(_szBuildTime+0);
            _cs.n8Hours10 = u8 >> 4;
            _cs.n8Hours1 = u8 & 0x0F;
        }

        RTCWriteTime(); // Copy RAM time to RTC
        
        // Put 2048Hz on CLKOUT pin (wired to T1CKI))
        I2CWriteRegister(I2C_RV3028C7_ADDR,0x0A,0x02); // Timer value 0
        I2CWriteRegister(I2C_RV3028C7_ADDR,0x0B,0x00); // Timer value 1
        I2CWriteRegister(I2C_RV3028C7_ADDR,0x0F,0b10011100); // Control 1 register; Periodic countdown enabled[TE-2], [TRPT-7]repeated, [TD-0:1] 4096Hz clock source, [USEL-4] set, [EERD-3] EEPROM daily refresh disabled
        I2CWriteRegister(I2C_RV3028C7_ADDR,0x10,0b00010000); // Control 2 register; [UIE-5] disabled, [TIE-4] enabled

        I2CWriteRegister(I2C_RV3028C7_ADDR,0x35,0xC6); // Periodic countdown on CLKOUT, wired to T1CKI
        
#ifdef ENABLE_VBAK_TRICKLE_CHARGE // Adds ~45nA @ 1.8V
        I2CWriteRegister(I2C_RV3028C7_ADDR,0x37,0x37); // TCE[5] on; FEDE[4] on (default); BSM[3:2]=0b01, DSM; TCR[0:1]=0b11 => 15k trickle charge; DSM adds 45nA @1.8V
#else        
        I2CWriteRegister(I2C_RV3028C7_ADDR,0x37,0x10); // TCE[5] off; FEDE[4] on (default); TCR[0:1]=0b00 => 3k trickle charge ~(default)
#endif
        
        I2CWriteRegister(I2C_RV3028C7_ADDR,0x0E,0x00); // Clear outstanding POR in RTC status register

        // Reset the first byte of HEF to zeros to indicate we can use the RTC for time next time if it still has valid time
        if (_au8HEF[0]==0xFF)
        {
            HEFClearFirstByte(); 
        }
    }
    else
    {
        RTCReadTime(); // If the RTC has valid time despite PIC reset, let's use it
    }
}

void BtnInit(void)
{
    // DOWN button on PORTB/IOC
    TRISBbits.TRISB1=1;
    WPUBbits.WPUB1=1;
    ANSELBbits.ANSB1=0;
    IOCBNbits.IOCBN1=1;
    IOCBPbits.IOCBP1=1;
    
    // UP button on PORTB/IOC
    TRISBbits.TRISB5=1;
    WPUBbits.WPUB5=1;
    ANSELBbits.ANSB5=0;
    IOCBNbits.IOCBN5=1;
    IOCBPbits.IOCBP5=1;
    
    IOCBF=0; // Clear any outstanding stray IOC notifications: this also clears IOCIF
    INTCONbits.IOCIE=1;
    
    // HOUR button on RE3
    // TRISEbits.TRISE3=1; // Always an input
    WPUEbits.WPUE3=1;
    
    LPDelayMS(1);
    
    _u8BtnPortBState=PORTB;
    _u8BtnPortBStateLast=_u8BtnPortBState;
    
}

static uint8_t BtnReadDebounced(void)
{
    // This debounces all bits on PORTB: it's cheaper than trying to do this one bit at a time.
    uint8_t u8PortBStateNow;
    uint8_t u8PortBStableMask;

    u8PortBStateNow=PORTB;
    u8PortBStableMask= u8PortBStateNow ^ _u8BtnPortBStateLast; // bit 1 means port has changed since last time
    _u8BtnPortBStateLast=u8PortBStateNow;
    _u8BtnPortBState= (_u8BtnPortBState & u8PortBStableMask) | (u8PortBStateNow & ~u8PortBStableMask); // maintain old state | merge stable state 
    IOCBF=0; // Clear any outstanding stray IOC notifications: this also clears IOCIF
    return _u8BtnPortBState;
}

static void BtnProcess(void)
{
    U8UNION u8uBtnState;
    static U8UNION u8uBtnStateLast={.u8=0xFF};
    bool bFinished=false;
    uint16_t u16Counter=0; // How many iterations with no button action before we return to normal service
    uint16_t u16ColonCounter=0; // Quick flash of colon while in setting mode

    LCD_DATA_COL=0;

    do
    {
        u8uBtnState.u8=BtnReadDebounced(); // Get the debounced state of PORTB
        {
            if (u8uBtnState.s.b1==0 && u8uBtnStateLast.s.b1==1) // Down button pushed
            {
                if (BTN_HOURS)
                {
                    if (--_cs.n8Mins1<0)
                    {
                        _cs.n8Mins1=9;
                        if (--_cs.n8Mins10<0)
                        {
                            _cs.n8Mins10=5;
                        }
                        LCDDigit3Set((uint8_t)_cs.n8Mins10);
                    }
                    LCDDigit4Set((uint8_t)_cs.n8Mins1);
                }
                else
                {
                    if (--_cs.n8Hours1<0)
                    {
                        _cs.n8Hours1=9;
                        if (--_cs.n8Hours10<0)
                        {
                            _cs.n8Hours10=2;
                            _cs.n8Hours1=3;
                        }
                        LCDDigit1Set((uint8_t)_cs.n8Hours10);
                    }
                    LCDDigit2Set((uint8_t)_cs.n8Hours1);
                }
            }
            if (u8uBtnState.s.b5==0 && u8uBtnStateLast.s.b5==1) // Up button pushed
            {
                if (BTN_HOURS)
                {
                    if (++_cs.n8Mins1>9)
                    {
                        _cs.n8Mins1=0;
                        if (++_cs.n8Mins10>5)
                        {
                            _cs.n8Mins10=0;
                        }
                        LCDDigit3Set((uint8_t)_cs.n8Mins10);
                    }
                    LCDDigit4Set((uint8_t)_cs.n8Mins1);
                }
                else
                {
                    if (++_cs.n8Hours1>9)
                    {
                        _cs.n8Hours1=0;
                        if (++_cs.n8Hours10>9)
                        {
                            _cs.n8Hours10=0;
                        }
                        LCDDigit1Set((uint8_t)_cs.n8Hours10);
                    }
                    if (_cs.n8Hours1>3 && _cs.n8Hours10>=2)
                    {
                        _cs.n8Hours1=0;
                        _cs.n8Hours10=0;
                        LCDDigit1Set((uint8_t)_cs.n8Hours10);
                    }                        
                    LCDDigit2Set((uint8_t)_cs.n8Hours1);
                }                    
            }
        }
        if (u8uBtnState.s.b1==0 || u8uBtnState.s.b5==0) // If either button down
        {
            u16Counter=0;
        }
        u8uBtnStateLast=u8uBtnState;
        if (++u16ColonCounter>50)
        {
            LCD_DATA_COL=!LCD_DATA_COL;
            u16ColonCounter=0;
        }

        if (++u16Counter>500)
        {
            bFinished=true;
        }
        else
        {
            LPDelayMS(1);
        }
    } while (!bFinished);

    LCD_DATA_COL=1;

    RTCWriteTime();

}

static void CPUInit(void)
{
    OSCCONbits.IRCF=0b1111; // 0b1111 => 16MHz
    
    // GPIO direction
    TRISA=0b00000000; 
    TRISB=0b00101010; // RB1+RB5 are IOC, RB3 is VLCD3
    TRISC=0b00000011; // RC0/T1CKI, RC1 is SCL
    TRISD=0b00000001; // RD0 is SDA
    TRISE=0b00001000; // RE3 is MCLR/RE3
    
    // Pull up settings for now
    WPUB=0;
    WPUE=0;
    OPTION_REGbits.nWPUEN=0; // Enable pull ups globally
}

static void T1OscInit(void) // Enable T1CKI, it's required for LCD, wired the RTC's CLKOUT.
{
    TRISCbits.TRISC0=1; // RC0/T1CKI. No analog function on this pin, so no ANSEL bit config to perform
    T1CONbits.T1OSCEN=0;   
}

static void T1Init(void) // Setup timer 1 for 1 second interrupt: T1CKI should already have 2048Hz on it. T1OscInit() should have been run before this.
{
    T1CONbits.TMR1CS=0b10; // 0b10 => T1CKI
//    T1CONbits.T1CKPS=0b00; // 0b00 => 1:1 prescale
    T1CONbits.T1CKPS=0b01; // 0b01 => 1:2 prescale => 1024Hz TMR1 increment from 2048Hz T1CLKI
    T1CONbits.nT1SYNC=1; // 1 => Not sync'd with Fcy 
    TMR1=(uint16_t)(65536UL-TIMER1_DELAY-1); // 1min timeout on timer 1
    PIR1bits.TMR1IF=0;
    PIE1bits.TMR1IE=1;
    T1CONbits.TMR1ON=1;
}

static void LCDUpdate(void)
{
    // Setup the display digits with the current RAM time
    LCDDigit1Set((uint8_t)_cs.n8Hours10);
    LCDDigit2Set((uint8_t)_cs.n8Hours1);
    LCDDigit3Set((uint8_t)_cs.n8Mins10);
    LCDDigit4Set((uint8_t)_cs.n8Mins1);    
}

int main(void) 
{
    CPUInit();
    BtnInit();
    I2CInit();

    LPDelayMS(50); // Low power delay, wait for buses to settle
    
    RTCInit();
    T1OscInit(); // Enable T1CKI, it's required for LCD, & is wired the RTC's CLKOUT running at 2048Hz
    LCDInit();
    LCDUpdate();
    T1Init(); // Setup timer 1 for 1 second interrupt: T1CKI should already have 2048Hz on it
    
    INTCONbits.PEIE=1; // Enable peripheral interrupts to allow wake from sleep: we don't actually use interrupt handlers, just the wake from sleep functionality.
    
    while(1)
    {
        LATAbits.LATA2=0; // GPIO for debug to measure time out of sleep

        SLEEP();

        LATAbits.LATA2=1; // GPIO for debug to measure time out of sleep
        
        if (INTCONbits.IOCIF) // Up or Down button has been pressed
        {
            BtnProcess();

            // Reset the timer.
            TMR1=(uint16_t)(65536UL-TIMER1_DELAY-1); // Reset the 1min timeout on timer 1
            PIR1bits.TMR1IF=0;
        }
    
        if (PIR1bits.TMR1IF)
        {
            TMR1-=TIMER1_DELAY; // We can do safely because TMR1 clock is only 1024Hz (after 1:2 prescaler), we're running @ 4MIPS, so we can get in plenty of instructions between waking from sleep and getting here
            LATAbits.LATA2=1; // GPIO for debug
            PIR1bits.TMR1IF=0;
            
            // Note that to minimise CPU utilisation & power consumption, we only do the absolute minimum cycles, and don't waste power asking the RTC for the time.
            if (++_cs.n8Mins1>9)
            {
                _cs.n8Mins1=0;
                if (++_cs.n8Mins10>5)
                {
                    _cs.n8Mins10=0;

                    if (++_cs.n8Hours1>9)
                    {
                        _cs.n8Hours1=0;
                        if (++_cs.n8Hours10>9)
                        {
                            _cs.n8Hours10=0;
                        }
                        LCDDigit1Set((uint8_t)_cs.n8Hours10);
                    }
                    if (_cs.n8Hours1>3 && _cs.n8Hours10>=2)
                    {
                        _cs.n8Hours1=0;
                        _cs.n8Hours10=0;
                        LCDDigit1Set((uint8_t)_cs.n8Hours10);
                    }
                    LCDDigit2Set((uint8_t)_cs.n8Hours1);
                }
                LCDDigit3Set((uint8_t)_cs.n8Mins10);
            }
            LCDDigit4Set((uint8_t)_cs.n8Mins1);
        }
    }    

    return 0;
}
